package server.api.vectordbs;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.Writer;

import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import org.eclipse.jetty.server.Request;
import org.eclipse.jetty.server.Response;
import org.eclipse.jetty.server.UserIdentity;
import org.tinylog.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import broker.Broker;
import pointcloud.Rect2d;
import util.IndentedXMLStreamWriter;
import util.Timer;
import util.Web;
import util.XmlUtil;
import vectordb.VectorDB;

public class VectordbHandler_wfs extends VectordbHandler {	

	private static final String FEATURE_NAME = "feature";
	private static final String GEOMETRY_PROPERTY_NAME = "geometry";

	public VectordbHandler_wfs(Broker broker) {
		super(broker, "wfs");
	}

	@Override
	public void handleGET(VectorDB vectordb, String target, Request request, Response response, UserIdentity userIdentity) throws IOException {
		request.setHandled(true);

		/*if(!"WFS".equals(request.getParameter("SERVICE"))) {
			Logger.error("no WFS");
			return;
		}*/		

		String reqParam = Web.getLastString(request, "REQUEST", null);
		if(reqParam == null) {
			reqParam = Web.getLastString(request, "Request", null);
		}
		if(reqParam == null) {
			reqParam = Web.getLastString(request, "request", null);
		}
		if(reqParam == null) {
			reqParam = "GetCapabilities";
		}		

		
		switch (reqParam) {
		case "GetCapabilities":
			handle_GetCapabilities(vectordb, request, response);
			break;
		case "DescribeFeatureType":
			handle_DescribeFeatureType(vectordb, request, response);
			break;
		case "GetFeature":
			Timer.start("GetFeature");
			handle_GetFeature(vectordb, null, request, response);
			Logger.info(Timer.stop("GetFeature"));
			break;			
		default:
			Logger.error("unknown request "+reqParam);
		}		
	}

	private static void handle_GetCapabilities(VectorDB vectordb, Request request, Response response) throws IOException {
		String requestUrl = request.getRequestURL().toString() + "?rsdb";
		response.setContentType(Web.MIME_XML);	
		PrintWriter out = response.getWriter();	
		try {
			//out.write(cc);
			XmlUtil.writeXML(out, doc -> xmlGetCapabilities(vectordb, doc, requestUrl));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	private static void handle_DescribeFeatureType(VectorDB vectordb, Request request, Response response) throws IOException {
		response.setContentType("text/xml; subtype=gml/3.1.1; charset=UTF-8");
		PrintWriter out = response.getWriter();	
		try {
			//out.write(dd);
			XmlUtil.writeXML(out, doc -> xmlDescribeFeatureType(vectordb, doc));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * 
	 * @param vectordb
	 * @param ext nullable
	 * @param request
	 * @param response
	 * @throws IOException
	 */
	public static void handle_GetFeature(VectorDB vectordb, Rect2d ext, Request request, Response response) throws IOException {
		response.setContentType("text/xml; subtype=gml/3.1.1; charset=UTF-8");
		PrintWriter out = response.getWriter();	
		try {
			//out.write(ff);
			//writeXML(out, doc -> xmlGetFeature(vectordb, doc));
			xmlGetFeatureStream(vectordb, ext, out);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}
	
	private static Node xmlGetCapabilities(VectorDB vectordb, Document doc, String requestUrl) {
		Element rootElement = doc.createElementNS("http://www.opengis.net/wfs", "WFS_Capabilities");
		rootElement.setAttribute("version", "1.1.0");
		rootElement.setAttribute("xmlns:xlink", "http://www.w3.org/1999/xlink");
		addOperationsMetadata(rootElement, requestUrl);
		addFeatureTypeList(rootElement, vectordb);
		return rootElement;
	}
	
	
	private static void addOperationsMetadata(Element rootElement, String requestUrl) {
		Element eOperationsMetadata = XmlUtil.addElement(rootElement, "OperationsMetadata");
		
		Element eOperationGetCapabilities = XmlUtil.addElement(eOperationsMetadata, "Operation");
		eOperationGetCapabilities.setAttribute("name", "GetCapabilities");
		Element eOperationGetCapabilitiesDCP = XmlUtil.addElement(eOperationGetCapabilities, "DCP");
		Element eOperationGetCapabilitiesHTTP = XmlUtil.addElement(eOperationGetCapabilitiesDCP, "HTTP");
		Element eOperationGetCapabilitiesGet = XmlUtil.addElement(eOperationGetCapabilitiesHTTP, "Get");
		eOperationGetCapabilitiesGet.setAttribute("xlink:href", requestUrl);
		
		Element eOperationDescribeFeatureType = XmlUtil.addElement(eOperationsMetadata, "Operation");
		eOperationDescribeFeatureType.setAttribute("name", "DescribeFeatureType");
		Element eOperationDescribeFeatureTypeDCP = XmlUtil.addElement(eOperationDescribeFeatureType, "DCP");
		Element eOperationDescribeFeatureTypeHTTP = XmlUtil.addElement(eOperationDescribeFeatureTypeDCP, "HTTP");
		Element eOperationDescribeFeatureTypeGet = XmlUtil.addElement(eOperationDescribeFeatureTypeHTTP, "Get");
		eOperationDescribeFeatureTypeGet.setAttribute("xlink:href", requestUrl);
		
		Element eOperationGetFeature = XmlUtil.addElement(eOperationsMetadata, "Operation");
		eOperationGetFeature.setAttribute("name", "GetFeature");
		Element eOperationGetFeatureDCP = XmlUtil.addElement(eOperationGetFeature, "DCP");
		Element eOperationGetFeatureHTTP = XmlUtil.addElement(eOperationGetFeatureDCP, "HTTP");
		Element eOperationGetFeatureGet = XmlUtil.addElement(eOperationGetFeatureHTTP, "Get");
		eOperationGetFeatureGet.setAttribute("xlink:href", requestUrl);
	}
	
	private static void addFeatureTypeList(Element rootElement, VectorDB vectordb) {
		Element eFeatureTypeList = XmlUtil.addElement(rootElement, "FeatureTypeList");
		Element eFeatureType = XmlUtil.addElement(eFeatureTypeList, "FeatureType");
		//addElement(eFeatureType, "Name", vectordb.getName());
		//addElement(eFeatureType, "Title", vectordb.getName());	
		XmlUtil.addElement(eFeatureType, "Name", FEATURE_NAME);
		XmlUtil.addElement(eFeatureType, "Title", vectordb.getName());

		String epsg = vectordb.getDetails().epsg;		
		//addElement(eFeatureType, "DefaultSRS", "urn:ogc:def:crs:EPSG::" + epsg);
		XmlUtil.addElement(eFeatureType, "DefaultSRS", "EPSG:" + epsg);
	}
	
	private static Node xmlDescribeFeatureType(VectorDB vectordb, Document doc) {
		Element rootElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "schema");
		rootElement.setAttribute("version", "0.1");
		rootElement.setAttribute("targetNamespace", "https://environmentalinformatics-marburg.github.io/rsdb");
		rootElement.setAttribute("xmlns:gml", "http://www.opengis.net/gml");

		Element eimport = XmlUtil.addElement(rootElement, "import");
		eimport.setAttribute("namespace", "http://www.opengis.net/gml");
		eimport.setAttribute("schemaLocation", "http://schemas.opengis.net/gml/3.1.1/base/gml.xsd");

		Element eelement = XmlUtil.addElement(rootElement, "element");		
		eelement.setAttribute("name", FEATURE_NAME);
		//eelement.setAttribute("name", vectordb.getName());
		eelement.setAttribute("substitutionGroup", "gml:_Feature");

		Element ecomplexType = XmlUtil.addElement(eelement, "complexType");
		Element ecomplexContent = XmlUtil.addElement(ecomplexType, "complexContent");
		Element eextension = XmlUtil.addElement(ecomplexContent, "extension");
		eextension.setAttribute("base", "gml:AbstractFeatureType");
		Element esequence = XmlUtil.addElement(eextension, "sequence");

		Element esequenceElement1 = XmlUtil.addElement(esequence, "element");
		esequenceElement1.setAttribute("name", GEOMETRY_PROPERTY_NAME);
		esequenceElement1.setAttribute("type", "gml:GeometryPropertyType");

		vectordb.forEachFieldDefinition((int layerIndex, int fieldIndex, String fieldName) -> {
			Element esequenceElement2 = XmlUtil.addElement(esequence, "element");
			esequenceElement2.setAttribute("name", fieldName);
			esequenceElement2.setAttribute("type", "string");
		});		

		return rootElement;
	}
	
	/**
	 * 
	 * @param vectordb
	 * @param ext nullable
	 * @param out
	 * @throws XMLStreamException
	 */
	private static void xmlGetFeatureStream(VectorDB vectordb, Rect2d ext, Writer out) throws XMLStreamException {
		XMLOutputFactory factory = XMLOutputFactory.newInstance();
		factory.setProperty("escapeCharacters", false); // needed for direct GML geometry output
		XMLStreamWriter xmlWriterInner = factory.createXMLStreamWriter(out);
		XMLStreamWriter xmlWriter = new IndentedXMLStreamWriter(xmlWriterInner);
		xmlWriter.writeStartDocument();
		xmlWriter.writeStartElement("wfs:FeatureCollection");
		xmlWriter.writeDefaultNamespace("https://environmentalinformatics-marburg.github.io/rsdb");
		xmlWriter.writeNamespace("wfs", "http://www.opengis.net/wfs");
		xmlWriter.writeNamespace("gml", "http://www.opengis.net/gml");
		
		vectordb.forEachFeature(ext, vectorFeature -> {			
			xmlWriter.writeStartElement("gml:featureMember");
			xmlWriter.writeStartElement(FEATURE_NAME);
			
			vectorFeature.forEachField(vectorFeatureField -> {
				String fieldName = vectorFeatureField.getName();
				String fieldValue = vectorFeatureField.getAsString();
				xmlWriter.writeStartElement(fieldName);
				xmlWriter.writeCharacters(XmlUtil.encodeXMLifNeeded(fieldValue));
				xmlWriter.writeEndElement(); // fieldName
			});
			
			xmlWriter.writeStartElement(GEOMETRY_PROPERTY_NAME);  // geometry after properties for better properties reading by qgis
			String gml = vectorFeature.getGeometryGML();
			xmlWriter.writeCharacters(gml);
			xmlWriter.writeEndElement(); // GEOMETRY_PROPERTY_NAME

			xmlWriter.writeEndElement(); // FEATURE_NAME
			xmlWriter.writeEndElement(); // gml:featureMember
		});

		xmlWriter.writeEndElement(); // wfs:FeatureCollection
		xmlWriter.writeEndDocument();
		xmlWriter.close();
	}	

	static final String cc = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
			+ "<wfs:WFS_Capabilities xmlns:gml=\"http://www.opengis.net/gml\" xmlns:wfs=\"http://www.opengis.net/wfs\" xmlns:ows=\"http://www.opengis.net/ows\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:ogc=\"http://www.opengis.net/ogc\" xmlns:dvg=\"http://www.geobasis.nrw.de\" xmlns=\"http://www.opengis.net/wfs\" version=\"1.1.0\" xsi:schemaLocation=\"http://www.opengis.net/wfs http://schemas.opengis.net/wfs/1.1.0/wfs.xsd\">\r\n"
			+ "  <ows:ServiceIdentification>\r\n"
			+ "    <ows:Title>WFS NW DVG</ows:Title>\r\n"
			+ "    <ows:Abstract>Die Digitalen Verwaltungsgrenzen</ows:Abstract>\r\n"
			+ "    <ows:ServiceType codeSpace=\"OGC\">OGC WFS</ows:ServiceType>\r\n"
			+ "    <ows:ServiceTypeVersion>1.1.0</ows:ServiceTypeVersion>\r\n"
			+ "  </ows:ServiceIdentification>\r\n"
			+ "  <ows:OperationsMetadata>\r\n"
			+ "    <ows:Operation name=\"GetCapabilities\">\r\n"
			+ "      <ows:DCP>\r\n"
			+ "        <ows:HTTP>\r\n"
			+ "          <ows:Get xlink:type=\"simple\" xlink:href=\"http://127.0.0.1:8081/vectordbs/plots_forest_edge/wfs?\"/>\r\n"
			+ "        </ows:HTTP>\r\n"
			+ "      </ows:DCP>\r\n"
			+ "      <ows:Parameter name=\"service\">\r\n"
			+ "        <ows:Value>WFS</ows:Value>\r\n"
			+ "      </ows:Parameter>\r\n"
			+ "      <ows:Parameter name=\"AcceptVersions\">\r\n"
			+ "        <ows:Value>1.1.0</ows:Value>\r\n"
			+ "      </ows:Parameter>\r\n"
			+ "      <ows:Parameter name=\"AcceptFormats\">\r\n"
			+ "        <ows:Value>text/xml</ows:Value>\r\n"
			+ "      </ows:Parameter>\r\n"
			+ "    </ows:Operation>\r\n"
			+ "    <ows:Operation name=\"DescribeFeatureType\">\r\n"
			+ "      <ows:DCP>\r\n"
			+ "        <ows:HTTP>\r\n"
			+ "          <ows:Get xlink:type=\"simple\" xlink:href=\"http://127.0.0.1:8081/vectordbs/plots_forest_edge/wfs?\"/>\r\n"
			+ "        </ows:HTTP>\r\n"
			+ "      </ows:DCP>\r\n"
			+ "      <ows:Parameter name=\"outputFormat\">\r\n"
			+ "        <ows:Value>XMLSCHEMA</ows:Value>\r\n"
			+ "        <ows:Value>text/xml; subtype=gml/3.1.1</ows:Value>\r\n"
			+ "      </ows:Parameter>\r\n"
			+ "    </ows:Operation>\r\n"
			+ "    <ows:Operation name=\"GetFeature\">\r\n"
			+ "      <ows:DCP>\r\n"
			+ "        <ows:HTTP>\r\n"
			+ "          <ows:Get xlink:type=\"simple\" xlink:href=\"http://127.0.0.1:8081/vectordbs/plots_forest_edge/wfs?\"/>\r\n"
			+ "        </ows:HTTP>\r\n"
			+ "      </ows:DCP>\r\n"
			+ "      <ows:Parameter name=\"resultType\">\r\n"
			+ "        <ows:Value>results</ows:Value>\r\n"
			+ "        <ows:Value>hits</ows:Value>\r\n"
			+ "      </ows:Parameter>\r\n"
			+ "      <ows:Parameter name=\"outputFormat\">\r\n"
			+ "        <ows:Value>text/xml; subtype=gml/3.1.1</ows:Value>\r\n"
			+ "      </ows:Parameter>\r\n"
			+ "    </ows:Operation>\r\n"
			+ "  </ows:OperationsMetadata>\r\n"
			+ "  <FeatureTypeList>\r\n"
			+ "    <Operations>\r\n"
			+ "      <Operation>Query</Operation>\r\n"
			+ "    </Operations>\r\n"
			+ "    <FeatureType>\r\n"
			+ "      <Name>dvg:nw_dvg1_bld</Name>\r\n"
			+ "      <Title>DVG1 Landesgrenze NRW</Title>\r\n"
			+ "      <Abstract>DVG1 Landesgrenze NRW</Abstract>\r\n"
			+ "      <DefaultSRS>urn:ogc:def:crs:EPSG::25832</DefaultSRS>\r\n"
			+ "      <OutputFormats>\r\n"
			+ "        <Format>text/xml; subtype=gml/3.1.1</Format>\r\n"
			+ "      </OutputFormats>\r\n"
			+ "      <ows:WGS84BoundingBox dimensions=\"2\">\r\n"
			+ "        <ows:LowerCorner>5.765110958993 50.3102602941677</ows:LowerCorner>\r\n"
			+ "        <ows:UpperCorner>9.46865479713457 52.5320006856262</ows:UpperCorner>\r\n"
			+ "      </ows:WGS84BoundingBox>\r\n"
			+ "    </FeatureType>\r\n"
			+ "  </FeatureTypeList>\r\n"
			+ "</wfs:WFS_Capabilities>";

	static final String dd = "<?xml version='1.0' encoding=\"UTF-8\" ?>\r\n"
			+ "<schema\r\n"
			+ "   targetNamespace=\"http://www.geobasis.nrw.de\" \r\n"
			+ "   xmlns:dvg=\"http://www.geobasis.nrw.de\" \r\n"
			+ "   xmlns:ogc=\"http://www.opengis.net/ogc\"\r\n"
			+ "   xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\"\r\n"
			+ "   xmlns=\"http://www.w3.org/2001/XMLSchema\"\r\n"
			+ "   xmlns:gml=\"http://www.opengis.net/gml\"\r\n"
			+ "   elementFormDefault=\"qualified\" version=\"0.1\" >\r\n"
			+ "\r\n"
			+ "  <import namespace=\"http://www.opengis.net/gml\"\r\n"
			+ "          schemaLocation=\"http://schemas.opengis.net/gml/3.1.1/base/gml.xsd\" />\r\n"
			+ "\r\n"
			+ "  <element name=\"nw_dvg1_bld\" \r\n"
			+ "           type=\"dvg:nw_dvg1_bldType\" \r\n"
			+ "           substitutionGroup=\"gml:_Feature\" />\r\n"
			+ "\r\n"
			+ "  <complexType name=\"nw_dvg1_bldType\">\r\n"
			+ "    <complexContent>\r\n"
			+ "      <extension base=\"gml:AbstractFeatureType\">\r\n"
			+ "        <sequence>\r\n"
			+ "          <element name=\"msGeometry\" type=\"gml:GeometryPropertyType\" minOccurs=\"0\" maxOccurs=\"1\"/>\r\n"
			+ "          <element name=\"ART\" minOccurs=\"0\" type=\"string\"/>\r\n"
			+ "          <element name=\"GN\" minOccurs=\"0\" type=\"string\"/>\r\n"
			+ "          <element name=\"KN\" minOccurs=\"0\" type=\"string\"/>\r\n"
			+ "          <element name=\"STAND\" minOccurs=\"0\" type=\"string\"/>\r\n"
			+ "        </sequence>\r\n"
			+ "      </extension>\r\n"
			+ "    </complexContent>\r\n"
			+ "  </complexType>\r\n"
			+ "\r\n"
			+ "</schema>\r\n"
			+ "";

	static final String ff = "<?xml version='1.0' encoding=\"UTF-8\" ?>\r\n"
			+ "<wfs:FeatureCollection\r\n"
			+ "   xmlns:dvg=\"http://www.geobasis.nrw.de\"\r\n"
			+ "   xmlns:gml=\"http://www.opengis.net/gml\"\r\n"
			+ "   xmlns:wfs=\"http://www.opengis.net/wfs\"\r\n"
			+ "   xmlns:ogc=\"http://www.opengis.net/ogc\"\r\n"
			+ "   xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\r\n"
			+ "   xsi:schemaLocation=\"http://www.geobasis.nrw.de https://www.wfs.nrw.de/geobasis/wfs_nw_dvg?SERVICE=WFS&amp;VERSION=1.1.0&amp;REQUEST=DescribeFeatureType&amp;TYPENAME=dvg:nw_dvg1_bld&amp;OUTPUTFORMAT=text/xml;%20subtype=gml/3.1.1  http://www.opengis.net/wfs http://schemas.opengis.net/wfs/1.1.0/wfs.xsd\">\r\n"
			+ "      <gml:boundedBy>\r\n"
			+ "      	<gml:Envelope srsName=\"EPSG:25832\">\r\n"
			+ "      		<gml:lowerCorner>280374.968000 5577679.220000</gml:lowerCorner>\r\n"
			+ "      		<gml:upperCorner>531791.200000 5820211.608000</gml:upperCorner>\r\n"
			+ "      	</gml:Envelope>\r\n"
			+ "      </gml:boundedBy>\r\n"
			+ "    <gml:featureMember>\r\n"
			+ "      <dvg:nw_dvg1_bld gml:id=\"nw_dvg1_bld.05000000\">\r\n"
			+ "        <gml:boundedBy>\r\n"
			+ "        	<gml:Envelope srsName=\"EPSG:25832\">\r\n"
			+ "        		<gml:lowerCorner>280374.968000 5577679.220000</gml:lowerCorner>\r\n"
			+ "        		<gml:upperCorner>531791.200000 5820211.608000</gml:upperCorner>\r\n"
			+ "        	</gml:Envelope>\r\n"
			+ "        </gml:boundedBy>\r\n"
			+ "        <dvg:msGeometry>\r\n"
			+ "          <gml:MultiSurface srsName=\"EPSG:25832\">\r\n"
			+ "            <gml:surfaceMember>\r\n"
			+ "              <gml:Polygon>\r\n"
			+ "                <gml:exterior>\r\n"
			+ "                  <gml:LinearRing>\r\n"
			+ "                    <gml:posList srsDimension=\"2\">298837.283000 5722344.564000 298837.913000 5722357.450000 298838.627000 5722369.056000 298839.730000 5722382.350000 298842.160000 5722407.223000 5672820.238000 296529.311000 5672819.707000 296528.459000 5672819.503000 296527.712000 5672819.576000 296526.863000 5672819.883000 296522.565000 5672822.246000 296521.222000 5672822.607000 296519.656000 5672822.778000 296513.046000 5672822.695000 296511.682000 5672823.048000 296510.715000 5672823.731000 296510.140000 5672824.614000 296509.733000 5672825.826000 296509.497000 5672827.358000 296509.436000 5672829.188000 296509.538000 5672836.591000 296520.337000 5672843.657000 296521.416000 5672857.826000 296525.366000 5672858.966000 296526.433000 5672867.931000 296522.744000 5672868.282000 296521.349000 5672868.439000 296507.065000 5672861.816000 296504.222000 5672851.021000 296503.287000 5672848.397000 296501.927000 5672845.999000 296500.171000 5672843.876000 296498.051000 5672842.065000 296495.423000 5672840.524000 296492.755000 5672839.608000 296490.265000 5672839.388000 296488.052000 5672839.870000 296483.374000 5672842.691000 296482.204000 5672843.126000 296481.083000 5672843.258000 296479.570000 5672843.028000 296475.993000 5672841.950000 296475.164000 5672841.880000 296474.527000 5672842.024000 296473.956000 5672842.472000 296473.465000 5672843.268000 296472.224000 5672846.923000 296470.876000 5672850.681000 296465.421000 5672842.096000 296462.548000 5672852.623000 296448.861000 5672848.377000 296443.516000 5672846.334000 296442.354000 5672846.164000 296441.473000 5672846.377000 296440.675000 5672847.135000 296440.047000 5672848.469000 296439.594000 5672850.366000 296439.332000 5672852.770000 296438.728000 5672862.504000 296436.307000 5672859.600000 296433.669000 5672855.412000 296432.265000 5672853.760000 296430.362000 5672852.311000 296428.165000 5672851.198000 296426.263000 5672850.712000 296425.554000 5672850.741000 296425.044000 5672850.952000 296424.615000 5672851.749000 296424.679000 5672853.158000 296426.473000 5672859.001000 296427.221000 5672861.825000 296427.469000 5672864.068000 296427.238000 5672865.911000 296426.516000 5672867.407000 296425.742000 5672868.247000 296424.745000 5672868.959000 296423.575000 5672869.506000 296422.292000 5672869.861000 296420.973000 5672870.006000 296419.655000 5672869.940000 296418.381000 5672869.666000 296417.180000 5672869.190000 296410.846000 5672866.045000 296239.601000 5672908.171000 296212.945000 5672914.765000 296107.856000 5672940.637000 296033.576000 5672958.925000 296020.147000 5672995.307000 295989.681000 5673077.729000 295912.156000 5673287.748000 295908.388000 5673297.955000 295907.399000 5673300.698000 295867.376000 5673408.934000 295835.099000 5673496.034000 295828.877000 5673512.804000 295819.864000 5673537.195000 295816.434000 5673546.344000 295803.185000 5673582.218000 295801.478000 5673586.993000 295690.878000 5673788.097000 295562.401000 5674022.062000 295451.067000 5674224.799000 295450.504000 5674240.437000 295441.878000 5674467.906000 295432.758000 5674712.712000 295422.084000 5674995.418000 295405.612000 5675435.111000 295396.799000 5675667.793000 295388.109000 5675898.568000 295384.741000 5676000.000000 295380.704000 5676121.584000 295380.085000 5676140.727000 295379.612000 5676153.659000 295379.514000 5676156.165000 295379.148000 5676166.790000 295378.123000 5676197.815000 295377.922000 5676203.909000 295377.679000 5676211.259000 295377.267000 5676223.724000 295376.552000 5676245.372000 295374.183000 5676317.237000 295362.437000 5676666.751000 295362.378000 5676667.754000 295359.907000 5676742.026000 295350.999000 5677006.643000 295343.400000 5677234.671000 295340.157000 5677331.697000 295330.588000 5677616.860000 295259.560000 5677634.977000 295258.968000 5677951.204000 295258.551000 5678252.109000 295258.548000 5678256.914000 295258.495000 5678275.233000 295257.979000 5678436.509000 295358.395000 5678520.290000 295385.853000 5678543.091000 295453.728000 5678599.674000 295634.401000 5678631.528000 295880.473000 5678625.076000 295881.472000 5678625.035000 296049.555000 5678620.669000 296172.192000 5678611.353000 296248.320000 5678593.128000 296274.389000 5678586.858000 296354.027000 5678558.980000 296472.204000 5678602.196000 296526.468000 5678622.100000 296479.330000 5678776.571000 296441.283000 5678901.243000 296431.745000 5678932.362000 296373.194000 5679123.535000 296366.206000 5679141.137000 296363.933000 5679146.735000 296292.238000 5679327.334000 296264.190000 5679397.847000 296258.291000 5679412.702000 296224.352000 5679498.269000 296220.635000 5679507.730000 296215.003000 5679521.773000 296205.111000 5679546.701000 296149.028000 5679688.025000 296122.253000 5679755.383000 296050.818000 5679934.971000 295998.165000 5680067.147000 295982.645000 5680106.218000 295934.806000 5680226.386000 295912.842000 5680281.636000 295861.573000 5680406.048000 295784.390000 5680604.189000 295752.703000 5680686.061000 295697.294000 5680829.161000 295681.361000 5680870.351000 295910.304000 5681000.726000 296142.626000 5681132.965000 296376.384000 5681266.045000 296585.958000 5681385.400000 296608.402000 5681398.195000 296791.322000 5681585.403000 296968.921000 5681767.322000 297119.046000 5681921.137000 297196.098000 5682000.000000 297362.139000 5682169.944000 297396.510000 5682205.173000 297591.449000 5682404.800000 297785.655000 5682603.656000 297899.948000 5682720.798000 297982.998000 5682805.786000 298178.353000 5683005.795000 298347.626000 5683179.945000 298519.725000 5683357.081000 298529.341000 5683366.986000 298673.019000 5683514.968000 298675.003000 5683516.989000 298809.161000 5683654.838000 298822.952000 5683668.988000 299022.004000 5683873.748000 299127.760000 5683982.830000 299236.753000 5684095.183000 299299.767000 5684160.169000 299435.405000 5684299.959000 299439.586000 5684304.292000 299453.803000 5684451.249000 299469.797000 5684617.151000 299520.295000 5684753.015000 299555.175000 5684847.178000 299635.481000 5685063.300000 299693.192000 5685218.887000 299746.740000 5685363.234000 299801.963000 5685511.817000 299952.800000 5685683.014000 300000.000000 5685736.634000 300056.302000 5685800.594000 300161.159000 5685919.520000 300279.982000 5686054.391000 300318.922000 5686098.541000 300431.451000 5686226.361000 300444.575000 5686241.339000 300495.483000 5686299.112000 300501.465000 5686305.874000 300549.341000 5686360.469000 300591.639000 5686408.385000 300654.515000 5686479.782000 300698.373000 5686529.135000 300729.227000 5686564.007000 300787.165000 5686629.700000 300805.011000 5686649.990000 300841.855000 5686691.823000 300919.328000 5686779.939000 300980.088000 5686848.520000 301011.555000 5686898.380000 301128.336000 5687083.679000 301308.138000 5687365.893000 301394.880000 5687501.974000 301468.540000 5687616.770000 301563.072000 5687764.644000 301624.142000 5687860.442000 301625.321000 5687862.290000 301626.021000 5687867.166000 301645.627000 5688000.000000 301652.469000 5688046.351000 301660.530000 5688101.473000 301661.517000 5688118.587000 301669.266000 5688253.013000 301670.184000 5688268.934000 301688.819000 5688290.031000 301716.259000 5688321.099000 301740.013000 5688363.804000 301765.849000 5688410.254000 301796.731000 5688447.952000 301824.637000 5688482.018000 301913.557000 5688590.584000 301914.884000 5688592.203000 302030.481000 5688733.288000 302050.833000 5688758.127000 302053.994000 5688806.863000 302057.896000 5688867.008000 302061.252000 5688918.743000 302065.474000 5688983.838000 302080.036000 5689207.650000 302227.585000 5689413.408000 302228.329000 5689414.479000 302268.460000 5689570.883000 302348.476000 5689711.341000 302421.586000 5689839.571000 302425.578000 5689846.614000 302513.013000 5689943.329000 302533.455000 5689965.914000 302553.689000 5689988.307000 302554.519000 5689989.223000 302595.661000 5690009.409000 302635.379000 5690057.131000 302800.474000 5690253.862000 302802.159000 5690255.895000 302803.627000 5690257.537000 302795.354000 5690314.628000 302789.490000 5690340.091000 302780.603000 5690365.078000 302773.965000 5690383.866000 302750.916000 5690429.651000 302748.418000 5690434.658000 302736.547000 5690482.087000 302734.141000 5690491.794000 302733.322000 5690512.034000 302732.101000 5690542.224000 302761.815000 5690661.720000 302765.679000 5690660.761000 302833.132000 5690667.014000 302846.534000 5690668.256000 302859.800000 5690669.485000 302958.186000 5690678.606000 302971.631000 5690676.955000 303150.478000 5690654.521000 303427.202000 5690683.133000 303437.256000 5690662.202000 303489.497000 5690686.489000 303484.800000 5690698.892000 303578.004000 5690728.709000 303641.441000 5690735.521000 303779.480000 5690750.490000 303810.315000 5690760.439000 303941.567000 5690802.811000 304068.560000 5690829.142000 304145.879000 5690840.091000 304326.993000 5690819.266000 304327.784000 5690819.034000 304426.381000 5690784.971000 304504.541000 5690789.579000 304548.132000 5690800.607000 304558.247000 5690803.196000 304541.373000 5690836.015000 304520.539000 5690864.994000 304495.162000 5690900.364000 304491.182000 5690905.832000 304426.701000 5691039.791000 304426.242000 5691040.810000 304395.911000 5691099.404000 304395.168000 5691100.836000 304385.957000 5691120.330000 304385.215000 5691121.762000 304383.460000 5691125.337000 304366.827000 5691159.148000 304359.795000 5691170.746000 304347.990000 5691190.447000 304347.427000 5691191.371000 304336.657000 5691207.026000 304302.164000 5691273.997000 304274.503000 5691324.475000 304276.475000 5691326.196000 304336.626000 5691377.436000 304378.834000 5691413.390000 304404.344000 5691435.066000 304409.221000 5691439.271000 304416.373000 5691445.284000 304422.905000 5691450.822000 304458.676000 5691481.087000 304466.972000 5691488.154000 304473.501000 5691493.592000 304492.582000 5691509.827000 304499.114000 5691515.364000 304574.183000 5691578.953000 304579.784000 5691583.728000 304584.968000 5691588.120000 304592.328000 5691594.325000 304600.451000 5691601.200000 304616.694000 5691614.947000 304619.078000 5691616.952000 304714.156000 5691697.437000 304723.485000 5691705.263000 304773.148000 5691747.270000 304918.405000 5691870.142000 305008.091000 5691946.043000 305026.028000 5691961.224000 305238.565000 5692140.795000 305245.613000 5692146.713000 305501.601000 5692363.240000 305673.604000 5692508.638000 305945.856000 5692738.612000 306000.000000 5692784.364000 306090.693000 5692860.999000 306192.083000 5692946.529000 306358.789000 5693087.238000 306723.007000 5693395.019000 306763.753000 5693429.484000 306771.114000 5693435.688000 306904.654000 5693548.628000 306853.998000 5693773.302000 306849.596000 5693792.900000 306847.165000 5693803.807000 306845.791000 5693809.971000 306803.061000 5694000.000000 306787.367000 5694069.796000 306786.941000 5694071.615000 306786.515000 5694073.434000 306747.972000 5694244.365000 306722.411000 5694358.314000 306691.089000 5694497.821000 306683.889000 5694529.744000 306660.153000 5694634.510000 306636.088000 5694740.991000 306595.463000 5694919.714000 306576.162000 5695005.682000 306542.473000 5695155.796000 306526.477000 5695227.115000 306516.829000 5695270.148000 306514.281000 5695281.263000 306503.622000 5695328.447000 306432.600000 5695642.832000 306421.336000 5695692.838000 306415.748000 5695717.689000 306410.845000 5695739.285000 306405.403000 5695763.253000 306400.896000 5695783.064000 306398.278000 5695794.573000 306394.043000 5695813.364000 306388.061000 5695839.842000 306377.885000 5695884.889000 306358.626000 5695970.155000 306356.254000 5695980.661000 306347.437000 5696019.557000 306325.243000 5696117.954000 306320.411000 5696139.271000 306308.283000 5696193.014000 306307.280000 5696197.461000 306305.330000 5696206.049000 306303.137000 5696216.048000 306301.358000 5696223.928000 306301.020000 5696225.443000 306293.719000 5696257.573000 306292.709000 5696262.016000 306292.374000 5696263.631000 306259.784000 5696403.892000 306225.836000 5696557.420000 306225.022000 5696559.555000 306210.812000 5696596.270000 306208.818000 5696601.356000 306298.712000 5696709.186000 306327.343000 5696743.535000 306197.231000 5696836.544000 306195.578000 5696837.712000 306408.442000 5697030.076000 306420.829000 5697041.182000 306427.693000 5697048.004000 306433.391000 5697054.671000 306439.215000 5697062.550000 306444.538000 5697070.544000 306456.516000 5697089.463000 306462.930000 5697098.921000 306470.081000 5697108.780000 306478.127000 5697119.262000 306487.649000 5697131.126000 306512.292000 5697160.945000 306539.830000 5697194.537000 306549.004000 5697205.136000 306557.522000 5697214.442000 306564.839000 5697221.999000 306581.750000 5697238.987000 306589.382000 5697246.968000 306599.740000 5697258.643000 306610.755000 5697272.090000 306621.206000 5697285.593000 306644.688000 5697316.768000 306652.040000 5697326.218000 306658.482000 5697334.140000 306664.578000 5697341.225000 306670.475000 5697347.640000 306682.174000 5697359.572000 306691.580000 5697369.808000 306701.201000 5697381.244000 306711.274000 5697394.149000 306722.764000 5697409.757000 306734.849000 5697426.764000 306754.629000 5697455.113000 306756.245000 5697457.429000 306761.348000 5697464.727000 306763.766000 5697467.301000 306851.255000 5697560.634000 306870.157000 5697582.280000 306874.818000 5697588.045000 306881.796000 5697597.618000 306889.297000 5697607.605000 306906.695000 5697629.939000 306914.373000 5697640.323000 306921.018000 5697650.001000 306929.624000 5697663.220000 306945.883000 5697688.248000 306957.888000 5697705.701000 306966.869000 5697717.909000 306989.433000 5697747.648000 307020.474000 5697789.516000 307033.185000 5697806.611000 307095.270000 5697887.943000 307074.377000 5697895.906000 307044.618000 5697907.334000 307091.944000 5697987.569000 306923.829000 5698005.566000 306873.085000 5698010.978000 306818.255000 5698016.837000 306391.075000 5698062.422000 306380.512000 5698063.555000 306376.627000 5698064.015000 306260.569000 5698076.416000 306068.096000 5698096.972000 306030.328000 5698101.021000 306010.344000 5698094.533000 306000.000000 5698091.181000 305985.391000 5698086.447000 305820.035000 5698032.566000 305817.092000 5698031.585000 305599.707000 5697961.018000 305669.758000 5698290.551000 305731.746000 5698582.079000 305751.084000 5698672.970000 305797.868000 5698892.895000 305820.630000 5698999.921000 305822.193000 5699007.284000 305825.158000 5699021.255000 305862.964000 5699199.419000 305883.902000 5699298.092000 305884.387000 5699300.384000 305944.812000 5699585.470000 305986.770000 5699781.631000 306000.000000 5699843.509000 306033.434000 5700000.000000 306038.954000 5700025.836000 306056.876000 5700109.759000 306057.374000 5700112.140000 306108.748000 5700352.435000 306110.000000 5700358.310000 306110.240000 5700359.301000 306111.160000 5700363.667000 306134.326000 5700473.218000 306159.014000 5700589.613000 306188.652000 5700729.126000 306214.568000 5700851.066000 306218.412000 5700869.135000 306238.477000 5700963.099000 306297.996000 5701243.116000 306299.647000 5701250.915000 306300.640000 5701255.609000 306335.009000 5701417.969000 306336.349000 5701424.310000 306336.559000 5701425.282000 306337.565000 5701430.045000 306353.887000 5701507.387000 306404.292000 5701746.220000 306406.412000 5701756.442000 306406.506000 5701802.680000 306406.472000 5701809.188000 306406.445000 5701818.297000 306405.837000 5702001.414000 306405.597000 5702073.563000 306405.020000 5702254.851000 306404.060000 5702578.184000 306403.416000 5702792.004000 306403.208000 5702840.657000 306402.995000 5702920.938000 306402.418000 5703129.051000 306402.362000 5703137.461000 306405.222000 5703143.749000 306424.515000 5703187.098000 306865.413000 5703160.217000 306863.079000 5703430.457000 306862.253000 5703730.162000 306861.518000 5704000.337000 306863.520000 5704000.355000 306861.761000 5704091.710000 306861.090000 5704129.071000 306864.953000 5704137.100000 306879.691000 5704167.744000 306892.063000 5704216.582000 306925.072000 5704346.947000 306945.951000 5704428.966000 306954.899000 5704464.132000 306974.682000 5704541.391000 306999.127000 5704637.177000 307011.314000 5704684.931000 307018.444000 5704712.854000 307026.448000 5704744.264000 307031.089000 5704762.481000 307040.262000 5704910.539000 307041.720000 5704933.900000 307042.066000 5704939.891000 307046.770000 5705015.567000 307048.492000 5705042.921000 307052.174000 5705103.425000 307053.019000 5705116.703000 307065.389000 5705315.876000 307065.982000 5705325.460000 307085.089000 5705366.715000 307105.480000 5705431.938000 307142.275000 5705538.427000 307144.324000 5705544.449000 307167.389000 5705594.349000 307172.656000 5705605.643000 307142.964000 5705655.305000 307118.131000 5705691.655000 307130.369000 5705770.425000 307155.431000 5705937.119000 307161.843000 5705979.745000 307164.889000 5706000.000000 307172.342000 5706049.558000 307174.286000 5706062.790000 307174.834000 5706071.275000 307181.257000 5706161.994000 307181.780000 5706169.880000 307182.279000 5706177.166000 307190.749000 5706300.730000 307098.865000 5706681.439000 307086.105000 5706711.990000 307081.690000 5706753.208000 307073.020000 5706780.988000 307060.013000 5706839.674000 307018.203000 5707013.645000 306943.126000 5707325.804000 306942.867000 5707326.815000 306921.571000 5707415.968000 306919.034000 5707426.582000 306880.431000 5707356.601000 306820.662000 5707490.471000 306799.857000 5707537.166000 306774.836000 5707593.342000 306774.278000 5707594.366000 306710.824000 5707704.266000 306693.141000 5707744.126000 306685.444000 5707761.457000 306662.446000 5707813.247000 306642.657000 5707857.898000 306619.930000 5707908.976000 306612.688000 5707925.187000 306610.949000 5707929.162000 306609.936000 5707931.306000 306579.050000 5707998.232000 306545.409000 5708071.177000 306510.755000 5708146.265000 306512.136000 5708160.421000 306515.653000 5708194.909000 306517.778000 5708268.689000 306520.540000 5708362.861000 306520.824000 5708367.353000 306522.358000 5708390.111000 306550.036000 5708569.540000 306559.101000 5708619.714000 306572.248000 5708710.858000 306573.351000 5708718.219000 306594.419000 5708863.387000 306616.135000 5709012.132000 306617.681000 5709022.979000 306609.769000 5709032.612000 306634.781000 5709234.869000 306658.770000 5709429.262000 306659.217000 5709432.847000 306643.246000 5709570.526000 306638.946000 5709607.235000 306638.787000 5709608.243000 306636.983000 5709649.654000 306635.334000 5709686.315000 306633.423000 5709728.772000 306632.272000 5709754.402000 306627.244000 5709866.049000 306622.129000 5709980.462000 306616.683000 5710082.313000 306610.689000 5710194.425000 306610.327000 5710200.245000 306586.552000 5710367.271000 306552.920000 5710603.864000 306551.552000 5710607.123000 306472.572000 5710792.328000 306364.748000 5711045.377000 306277.738000 5711249.530000 306187.187000 5711418.595000 306183.961000 5711424.623000 306026.821000 5711718.142000 306000.000000 5711768.140000 305986.469000 5711793.364000 305985.352000 5711795.412000 305934.962000 5711889.464000 305878.903000 5711994.157000 305878.344000 5711995.181000 305875.763000 5712000.000000 305818.380000 5712107.141000 305776.489000 5712185.370000 305773.384000 5712191.163000 305622.179000 5712338.296000 305555.429000 5712403.212000 305552.013000 5712406.535000 305411.594000 5712528.704000 305404.649000 5712535.095000 305351.694000 5712583.710000 305299.027000 5712632.014000 305265.647000 5712663.793000 305257.606000 5712671.448000 305235.643000 5712692.368000 305194.993000 5712731.070000 305153.960000 5712770.188000 305059.470000 5712860.345000 305030.763000 5712887.647000 305018.915000 5712898.943000 304959.593000 5712943.617000 304958.427000 5712944.465000 304688.904000 5713147.341000 304675.597000 5713157.358000 304628.704000 5713192.713000 304555.901000 5713247.100000 304526.345000 5713269.180000 304389.706000 5713371.277000 304388.444000 5713372.229000 304274.144000 5713457.695000 304224.614000 5713494.660000 304135.111000 5713720.938000 304053.935000 5713925.755000 304038.340000 5713965.231000 304035.805000 5713971.741000 303939.672000 5714214.507000 303936.678000 5714222.036000 303801.186000 5714564.208000 303697.260000 5714826.812000 303694.262000 5714834.242000 303691.547000 5714841.260000 303588.044000 5715101.945000 303468.028000 5715404.146000 303466.940000 5715406.893000 303455.418000 5715435.893000 303423.227000 5715517.288000 303323.619000 5715768.005000 303242.887000 5715971.404000 303157.670000 5716184.997000 303133.165000 5716246.359000 303111.019000 5716301.819000 303057.192000 5716436.452000 303043.477000 5716470.756000 303032.866000 5716497.306000 302976.233000 5716639.160000 302967.789000 5716660.226000 302966.067000 5716664.601000 302688.746000 5716858.057000 302563.612000 5716945.371000 302444.513000 5717028.534000 302316.214000 5717116.780000 302302.953000 5717126.170000 302292.118000 5717134.485000 302283.369000 5717141.887000 302275.510000 5717149.280000 302064.152000 5717359.847000 302063.672000 5717360.367000 301875.880000 5717550.048000 301783.237000 5717658.350000 301702.438000 5717752.853000 301671.587000 5717788.852000 301638.731000 5717827.236000 301500.851000 5717986.842000 301490.693000 5718000.000000 301463.026000 5718035.840000 301437.746000 5718068.607000 301346.534000 5718172.747000 301196.003000 5718350.890000 301065.202000 5718653.640000 300930.105000 5718958.970000 300908.580000 5719007.599000 300906.082000 5719015.008000 300904.751000 5719019.166000 300794.409000 5719268.946000 300792.841000 5719272.496000 300714.168000 5719452.103000 300711.853000 5719457.393000 300710.023000 5719461.572000 300690.540000 5719506.313000 300619.541000 5719668.780000 300470.983000 5719777.979000 300469.303000 5719779.048000 300467.393000 5719779.954000 300465.236000 5719780.705000 300462.784000 5719781.318000 300453.756000 5719782.590000 300345.131000 5719796.661000 300336.637000 5719797.740000 300331.597000 5719798.662000 300327.949000 5719799.649000 300324.751000 5719800.857000 300321.944000 5719802.310000 300319.509000 5719804.019000 300172.106000 5719924.282000 300169.228000 5719926.602000 300167.261000 5719928.184000 300000.000000 5720064.575000 299971.430000 5720087.872000 299964.941000 5720093.143000 299956.997000 5720099.575000 299817.033000 5720213.326000 299815.680000 5720214.483000 299657.026000 5720343.516000 299655.476000 5720344.781000 299510.782000 5720460.329000 299450.787000 5720508.136000 299442.737000 5720514.727000 299434.721000 5720521.509000 299416.730000 5720536.761000 299395.128000 5720554.265000 299389.264000 5720559.394000 299383.121000 5720565.049000 299369.012000 5720578.860000 299363.726000 5720584.182000 299356.719000 5720591.075000 299349.362000 5720597.584000 299279.516000 5720654.104000 299206.275000 5720713.267000 299196.973000 5720720.756000 299063.087000 5720828.953000 299059.026000 5720832.423000 298994.430000 5720887.627000 298818.890000 5721029.567000 298812.204000 5721034.947000 298595.284000 5721210.017000 298593.978000 5721211.557000 298593.193000 5721213.657000 298592.909000 5721216.326000 298593.088000 5721219.917000 298593.522000 5721223.446000 298594.154000 5721226.379000 298598.874000 5721243.802000 298622.127000 5721329.829000 298659.638000 5721504.656000 298661.416000 5721511.390000 298698.958000 5721655.286000 298731.278000 5721779.277000 298749.566000 5721849.193000 298758.281000 5721881.464000 298770.012000 5721922.924000 298772.986000 5721933.505000 298774.952000 5721941.232000 298776.460000 5721948.265000 298777.525000 5721954.746000 298778.306000 5721961.943000 298779.546000 5721980.329000 298780.424000 5721988.759000 298781.215000 5721993.765000 298782.246000 5721998.900000 298785.053000 5722009.664000 298788.897000 5722021.290000 298794.128000 5722034.840000 298797.795000 5722043.698000 298801.688000 5722053.164000 298804.173000 5722059.958000 298806.032000 5722066.215000 298807.268000 5722072.036000 298819.906000 5722148.190000 298821.229000 5722156.043000 298834.717000 5722216.347000 298838.561000 5722231.849000 298839.270000 5722236.331000 298839.508000 5722240.373000 298838.781000 5722251.914000 298837.046000 5722329.059000 298837.144000 5722341.166000 298837.283000 5722344.564000 </gml:posList>\r\n"
			+ "                  </gml:LinearRing>\r\n"
			+ "                </gml:exterior>\r\n"
			+ "              </gml:Polygon>\r\n"
			+ "            </gml:surfaceMember>            \r\n"
			+ "          </gml:MultiSurface>\r\n"
			+ "        </dvg:msGeometry>\r\n"
			+ "        <dvg:ART>L</dvg:ART>\r\n"
			+ "        <dvg:GN>Nordrhein-Westfalen</dvg:GN>\r\n"
			+ "        <dvg:KN>05000000</dvg:KN>\r\n"
			+ "        <dvg:STAND>20200304</dvg:STAND>\r\n"
			+ "      </dvg:nw_dvg1_bld>\r\n"
			+ "    </gml:featureMember>\r\n"
			+ "</wfs:FeatureCollection>";
}