package voxeldb;

import java.io.IOException;
import java.nio.file.Path;
import java.util.Arrays;


import org.tinylog.Logger;
import org.locationtech.proj4j.CRSFactory;
import org.locationtech.proj4j.CoordinateReferenceSystem;

import broker.TimeSlice;
import pointcloud.CellTable;
import pointcloud.Rect2d;
import pointdb.las.Las;
import pointdb.las.Laz;
import remotetask.CancelableRemoteProxy;
import util.Timer;
import util.Util;

public class Importer extends CancelableRemoteProxy {
	

	private static final CRSFactory CRS_FACTORY = new CRSFactory();

	private static final int READ_MAX_BYTES = 1_000_000_000;

	private final VoxelDB voxeldb;
	private final Rect2d filterRect;
	private final double filterZmin;
	private final double filterZmax;
	private final boolean trySetOriginToFileOrigin;
	private final TimeSlice timeSlice;
	private final CellFactory cellFactory;

	public Importer(VoxelDB voxeldb, Rect2d filterRect, double filterZmin, double filterZmax, boolean trySetOriginToFileOrigin, TimeSlice timeSlice) {
		this.voxeldb = voxeldb;
		this.filterRect = filterRect;
		this.trySetOriginToFileOrigin = trySetOriginToFileOrigin;
		this.timeSlice = timeSlice;
		this.cellFactory = CellFactory.ofAll(voxeldb);
		this.filterZmin = filterZmin;
		this.filterZmax = filterZmax;
	}

	/**
	 * 
	 * @param root directory or file
	 * @throws IOException
	 */
	public void importDirectory(Path root) throws IOException {
		Path[] paths = null;
		if(root.toFile().isFile()) {
			paths = new Path[]{root};
		} else {
			paths = Util.getPaths(root);
			for(Path path:paths) {
				if(isCanceled()) {
					throw new RuntimeException("canceled");
				}
				if(path.toFile().isDirectory()) {					
					importDirectory(path);
				}
			}
		}
		for(Path path:paths) {
			if(isCanceled()) {
				throw new RuntimeException("canceled");
			}
			if(path.toFile().isFile()) {				
				try {
					String filename = path.getFileName().toString().toLowerCase();
					String ext = filename.substring(filename.lastIndexOf('.')+1);
					if(ext.trim().toLowerCase().equals("las") || ext.trim().toLowerCase().equals("laz")) {
						//Logger.info("import file "+path);					
						importFile(path);
					} else {
						//Logger.info("skip file "+path);	
					}
				} catch(Exception e) {
					e.printStackTrace();
					Logger.error(e);
				}

			}
		}
	}

	private void importFile(Path filename) throws IOException {
		Logger.info("import " + filename);
		Timer.start("import");
		
		voxeldb.invalidateLocalRange(false);

		Las las = null;
		Laz laz = null;
		boolean isLas = true;
		int fileEPSG = 0;
		if(filename.toString().toLowerCase().endsWith("las")) {
			las = new Las(filename);
			Logger.info(las);			
			fileEPSG = las.readEPSG();			
		} else if(filename.toString().toLowerCase().endsWith("laz")) {
			laz = new Laz(filename);
			isLas = false;
			Logger.info(laz);
			fileEPSG = laz.readEPSG();
		} else {
			throw new RuntimeException("unknown extension");
		}

		if(fileEPSG != 0) {
			Logger.info("fileEpsg " + fileEPSG);
			if(voxeldb.geoRef().hasEpsg()) {
				//TODO
			} else {
				voxeldb.setEpsg(fileEPSG);
				if(!voxeldb.geoRef().hasProj4()) {
					try {
						String fileCode = "EPSG:"+fileEPSG;
						CoordinateReferenceSystem crs = CRS_FACTORY.createFromName(fileCode);
						if(crs != null) {
							String fileProj4 = crs.getParameterString();
							if(fileProj4 != null && !fileProj4.isEmpty()) {
								Logger.info("set proj4: " + fileProj4);
								voxeldb.setProj4(fileProj4);
							} else {
								Logger.warn("could not get proj4 of epsg");
							}
						} else {
							Logger.warn("could not get proj4 of epsg");
						}
					} catch(Exception e) {
						Logger.warn("could not get proj4 of epsg: " + e);
					}
				}
			}
		}

		double[] las_scale_factor = isLas ? las.scale_factor : laz.scale_factor;
		double[] las_offset = isLas ? las.offset : laz.offset;
		long las_number_of_point_records = isLas ? las.number_of_point_records : laz.number_of_point_records;
		final int record_count_max = READ_MAX_BYTES / (isLas ? las.point_Data_Record_Length : laz.point_Data_Record_Length);

		int xVoxelCellsize = voxeldb.getCellsize();
		int yVoxelCellsize = voxeldb.getCellsize();
		int zVoxelCellsize = voxeldb.getCellsize();

		double voxelSizeX = voxeldb.geoRef().voxelSizeX;
		double voxelSizeY = voxeldb.geoRef().voxelSizeY;
		double voxelSizeZ = voxeldb.geoRef().voxelSizeZ;

		double xProjectedCellsize = xVoxelCellsize * voxelSizeX;
		double yProjectedCellsize = yVoxelCellsize * voxelSizeY;
		double zProjectedCellsize = zVoxelCellsize * voxelSizeZ;

		double xlasmin = isLas ? las.min[0] : laz.min[0];
		double ylasmin = isLas ? las.min[1] : laz.min[1];
		double zlasmin = isLas ? las.min[2] : laz.min[2];
		double xlasmax = isLas ? las.max[0] : laz.max[0];
		double ylasmax = isLas ? las.max[1] : laz.max[1];
		double zlasmax = isLas ? las.max[2] : laz.max[2];

		if(filterRect != null) {
			if(xlasmin < filterRect.xmin) {
				xlasmin = filterRect.xmin;
			}
			if(ylasmin < filterRect.ymin) {
				ylasmin = filterRect.ymin;
			}
			if(xlasmax > filterRect.xmax) {
				xlasmax = filterRect.xmax;
			}
			if(ylasmax > filterRect.ymax) {
				ylasmax = filterRect.ymax;
			}
		}
		
		if(Double.isFinite(filterZmin) && zlasmin < filterZmin) {
			zlasmin = filterZmin;
		}
		
		if(Double.isFinite(filterZmax) && zlasmax > filterZmax) {
			zlasmax = filterZmax;
		}

		if(trySetOriginToFileOrigin) {
			voxeldb.trySetOrigin(xlasmin, ylasmin, zlasmin);
		}

		double xProjectedOrigin = voxeldb.geoRef().originX;
		double yProjectedOrigin = voxeldb.geoRef().originY;
		double zProjectedOrigin = voxeldb.geoRef().originZ;

		int xlascellmin = (int) Math.floor((xlasmin - xProjectedOrigin) / xProjectedCellsize);
		int xlascellmax = (int) Math.floor((xlasmax - xProjectedOrigin) / xProjectedCellsize);		
		int ylascellmin = (int) Math.floor((ylasmin - yProjectedOrigin) / yProjectedCellsize);
		int ylascellmax = (int) Math.floor((ylasmax - yProjectedOrigin) / yProjectedCellsize);		
		int zlascellmin = (int) Math.floor((zlasmin - zProjectedOrigin) / zProjectedCellsize);
		int zlascellmax = (int) Math.floor((zlasmax - zProjectedOrigin) / zProjectedCellsize);

		final int t = timeSlice.id;

		long pos = 0;
		int subdividedLen = 0;  // 0 == no subdivision
		int round = 0;
		while(pos < las_number_of_point_records) {
			round++;
			long current_long_len = subdividedLen > 0 ? subdividedLen : las_number_of_point_records - pos;
			int len = current_long_len >= record_count_max ? record_count_max : (int) current_long_len;
			setMessage("read records at " + pos + " len " + len + "  of " + las_number_of_point_records + "   round " + round);
			Timer.resume("get records");
			if(isCanceled()) {
				throw new RuntimeException("canceled");
			}
			CellTable recordTable = isLas ? las.getRecords(pos, len) : laz.getRecords(pos, len);
			if(isCanceled()) {
				throw new RuntimeException("canceled");
			}
			//Logger.info(Timer.stop("get records"));
			setMessage("process records at " + pos + " len " + len + "  of " + las_number_of_point_records + "   round " + round);

			Timer.resume("convert records");

			int[] xs = recordTable.x;
			int[] ys = recordTable.y;
			int[] zs = recordTable.z;
			char[] reds = recordTable.red;
			char[] greens = recordTable.green;
			char[] blues = recordTable.blue;

			int[] xrange = Util.getRange(xs);
			int[] yrange = Util.getRange(ys);
			int[] zrange = Util.getRange(zs);

			Logger.info("block range "+Arrays.toString(xrange) + " " + Arrays.toString(yrange) + " " + Arrays.toString(zrange));

			double xscale = las_scale_factor[0];
			double yscale = las_scale_factor[1];
			double zscale = las_scale_factor[2];
			double xoff = las_offset[0];
			double yoff = las_offset[1];
			double zoff = las_offset[2];

			double xmin = (xrange[0] * xscale) + xoff;
			double xmax = (xrange[1] * xscale) + xoff;
			double ymin = (yrange[0] * yscale) + yoff;
			double ymax = (yrange[1] * yscale) + yoff;			
			double zmin = (zrange[0] * zscale) + zoff;
			double zmax = (zrange[1] * zscale) + zoff;

			Logger.info("projected range "+xmin+" "+ymin+" "+zmin+"   "+xmax+" "+ymax+" "+zmax+"      "+(xmax-xmin)+" "+(ymax-ymin)+" "+(zmax-zmin));

			int xcellmin = (int) Math.floor((xmin - xProjectedOrigin) / xProjectedCellsize);
			int xcellmax = (int) Math.floor((xmax - xProjectedOrigin) / xProjectedCellsize);		
			int ycellmin = (int) Math.floor((ymin - yProjectedOrigin) / yProjectedCellsize);
			int ycellmax = (int) Math.floor((ymax - yProjectedOrigin) / yProjectedCellsize);			
			int zcellmin = (int) Math.floor((zmin - zProjectedOrigin) / zProjectedCellsize);
			int zcellmax = (int) Math.floor((zmax - zProjectedOrigin) / zProjectedCellsize);

			if(xcellmin < xlascellmin) {
				xcellmin = xlascellmin;
			}
			if(xcellmin > xlascellmax) {
				xcellmin = xlascellmax;
			}			
			if(ycellmin < ylascellmin) {
				ycellmin = ylascellmin;
			}
			if(ycellmin > ylascellmax) {
				ycellmin = ylascellmax;
			}			
			if(zcellmin < zlascellmin) {
				zcellmin = zlascellmin;
			}
			if(zcellmin > zlascellmax) {
				zcellmin = zlascellmax;
			}			
			if(xcellmax > xlascellmax) {
				xcellmax = xlascellmax;
			}
			if(xcellmax < xlascellmin) {
				xcellmax = xlascellmin;
			}			
			if(ycellmax > ylascellmax) {
				ycellmax = ylascellmax;
			}
			if(ycellmax < ylascellmin) {
				ycellmax = ylascellmin;
			}			
			if(zcellmax > zlascellmax) {
				zcellmax = zlascellmax;
			}
			if(zcellmax < zlascellmin) {
				zcellmax = zlascellmin;
			}

			int xcellrange = xcellmax - xcellmin + 1;
			int ycellrange = ycellmax - ycellmin + 1;
			int zcellrange = zcellmax - zcellmin + 1;
			long totalCellCount = ((long) xcellrange) * ((long) ycellrange) * ((long) zcellrange);
			Logger.info("cell "+xcellmin+" "+ycellmin+" "+zcellmin+"   "+xcellmax+" "+ycellmax+" "+zcellmax+"     "+xcellrange+" "+ycellrange+" "+zcellrange + "   " + totalCellCount);
			int maxCellCountPerBatch = 1024*1024;
			//long maxVoxels = 10_000_000_000l;
			//int maxCellCountPerBatch = (int) (((maxVoxels / xVoxelCellsize) / yVoxelCellsize) / zVoxelCellsize);
			if(totalCellCount > maxCellCountPerBatch) {
			subdividedLen = len / 2;
				if(subdividedLen < 1) {
					subdividedLen = 1;
				}
				Logger.info("point range to large: " + totalCellCount + "  of max " + maxCellCountPerBatch + " -> subdivide, slows down processing.  " + subdividedLen);
				continue; // process smaller set of point records	
			}
			subdividedLen = 0; // no subdivision
			pos += len; // process len point records

			Logger.info("VoxelCell[][][] cells...");
			VoxelCell[][][] cells = new VoxelCell[ycellrange][xcellrange][zcellrange];
			Logger.info("VoxelCell[][][] cells done.");

			for (int i = 0; i < len; i++) {
				double x = (xs[i] * xscale) + xoff;
				double y = (ys[i] * yscale) + yoff;
				double z = (zs[i] * zscale) + zoff;
				if(xlasmin <= x && x <= xlasmax && ylasmin <= y && y <= ylasmax && zlasmin <= z && z <= zlasmax) { // extent check of las file and filter rect					
					int xcell = (int) Math.floor((x - xProjectedOrigin) / xProjectedCellsize) - xcellmin;
					int ycell = (int) Math.floor((y - yProjectedOrigin) / yProjectedCellsize) - ycellmin;
					int zcell = (int) Math.floor((z - zProjectedOrigin) / zProjectedCellsize) - zcellmin;

					//Logger.info("point cell " + xcell + " " + ycell + " " + zcell);

					VoxelCell voxelCell = cells[ycell][xcell][zcell];
					if(voxelCell == null) {
						voxelCell = cellFactory.getVoxelCell(xcellmin + xcell, ycellmin + ycell, zcellmin + zcell, t);
						if(voxelCell == null) {
							voxelCell = VoxelCell.ofFilled(xcellmin + xcell, ycellmin + ycell, zcellmin + zcell,
									xVoxelCellsize, yVoxelCellsize, zVoxelCellsize,
									true, reds != null, greens != null, blues != null);
						} else {
							voxelCell = VoxelCell.ofFilled(voxelCell,
									xVoxelCellsize, yVoxelCellsize, zVoxelCellsize,
									true, reds != null, greens != null, blues != null);
						}
						cells[ycell][xcell][zcell] = voxelCell;
					}

					int xloc = (int) (floorMod(x - xProjectedOrigin, xProjectedCellsize) / voxelSizeX);
					int yloc = (int) (floorMod(y - yProjectedOrigin, yProjectedCellsize) / voxelSizeY);
					int zloc = (int) (floorMod(z - zProjectedOrigin, zProjectedCellsize) / voxelSizeZ);

					voxelCell.cnt[zloc][yloc][xloc]++;
					if(reds != null) {
						voxelCell.red[zloc][yloc][xloc] += reds[i];
					}
					if(greens != null) {
						voxelCell.green[zloc][yloc][xloc] += greens[i];
					}
					if(blues != null) {
						voxelCell.blue[zloc][yloc][xloc] += blues[i];
					}
				}
			}
			
			setMessage("write records at " + pos + " len " + len + "  of " + las_number_of_point_records + "   round " + round);

			for (int y = 0; y < ycellrange; y++) {
				VoxelCell[][] cellsY = cells[y];
				for (int x = 0; x < xcellrange; x++) {
					VoxelCell[] cellsYX = cellsY[x];				
					for (int z = 0; z < zcellrange; z++) {
						VoxelCell voxelCell = cellsYX[z];
						if(voxelCell != null) {
							cellFactory.writeVoxelCell(voxelCell, t);
						}
					}
				}
			}
			voxeldb.commit();
		}
		voxeldb.invalidateLocalRange(true);
		Logger.info(Timer.stop("import"));	
	}

	public static double floorMod(double x, double y) {
		return x - Math.floor(x / y) * y;
	}

	@Override
	public void process() throws Exception {
		throw new RuntimeException("not implemented: use importDirectory or importFile");
		
	}
}
